# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: kernel.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass

import betterproto


@dataclass(eq=False, repr=False)
class AccountAddress(betterproto.Message):
    """An address of an account. Always 32 bytes."""

    value: bytes = betterproto.bytes_field(1)


@dataclass(eq=False, repr=False)
class Memo(betterproto.Message):
    """
    A memo which can be included as part of a transfer. Max size is 256 bytes.
    This message can occur starting from protocol version 2.
    """

    value: bytes = betterproto.bytes_field(1)
