# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protocol-level-tokens.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import ccdexplorer_fundamentals.GRPCClient.kernel_pb2 as kernel__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bprotocol-level-tokens.proto\x12\x11\x63oncordium.v2.plt\x1a\x0ckernel.proto"\x15\n\x04\x43\x62or\x12\r\n\x05value\x18\x01 \x01(\x0c"\x18\n\x07TokenId\x12\r\n\x05value\x18\x01 \x01(\t"\x1f\n\x0eTokenModuleRef\x12\r\n\x05value\x18\x01 \x01(\x0c".\n\x0bTokenAmount\x12\r\n\x05value\x18\x01 \x01(\x04\x12\x10\n\x08\x64\x65\x63imals\x18\x02 \x01(\r"\xc6\x01\n\nTokenState\x12;\n\x10token_module_ref\x18\x01 \x01(\x0b\x32!.concordium.v2.plt.TokenModuleRef\x12\x10\n\x08\x64\x65\x63imals\x18\x03 \x01(\r\x12\x34\n\x0ctotal_supply\x18\x04 \x01(\x0b\x32\x1e.concordium.v2.plt.TokenAmount\x12-\n\x0cmodule_state\x18\x05 \x01(\x0b\x32\x17.concordium.v2.plt.CborJ\x04\x08\x02\x10\x03"\x95\x01\n\x11TokenAccountState\x12/\n\x07\x62\x61lance\x18\x01 \x01(\x0b\x32\x1e.concordium.v2.plt.TokenAmount\x12\x32\n\x0cmodule_state\x18\x04 \x01(\x0b\x32\x17.concordium.v2.plt.CborH\x00\x88\x01\x01\x42\x0f\n\r_module_stateJ\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04"J\n\x10TokenModuleEvent\x12\x0c\n\x04type\x18\x01 \x01(\t\x12(\n\x07\x64\x65tails\x18\x02 \x01(\x0b\x32\x17.concordium.v2.plt.Cbor"J\n\x0bTokenHolder\x12\x30\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1d.concordium.v2.AccountAddressH\x00\x42\t\n\x07\x61\x64\x64ress"\xcf\x01\n\x12TokenTransferEvent\x12,\n\x04\x66rom\x18\x01 \x01(\x0b\x32\x1e.concordium.v2.plt.TokenHolder\x12*\n\x02to\x18\x02 \x01(\x0b\x32\x1e.concordium.v2.plt.TokenHolder\x12.\n\x06\x61mount\x18\x03 \x01(\x0b\x32\x1e.concordium.v2.plt.TokenAmount\x12&\n\x04memo\x18\x04 \x01(\x0b\x32\x13.concordium.v2.MemoH\x00\x88\x01\x01\x42\x07\n\x05_memo"x\n\x16TokenSupplyUpdateEvent\x12.\n\x06target\x18\x01 \x01(\x0b\x32\x1e.concordium.v2.plt.TokenHolder\x12.\n\x06\x61mount\x18\x02 \x01(\x0b\x32\x1e.concordium.v2.plt.TokenAmount"\xc3\x02\n\nTokenEvent\x12,\n\x08token_id\x18\x01 \x01(\x0b\x32\x1a.concordium.v2.plt.TokenId\x12;\n\x0cmodule_event\x18\x02 \x01(\x0b\x32#.concordium.v2.plt.TokenModuleEventH\x00\x12?\n\x0etransfer_event\x18\x03 \x01(\x0b\x32%.concordium.v2.plt.TokenTransferEventH\x00\x12?\n\nmint_event\x18\x04 \x01(\x0b\x32).concordium.v2.plt.TokenSupplyUpdateEventH\x00\x12?\n\nburn_event\x18\x05 \x01(\x0b\x32).concordium.v2.plt.TokenSupplyUpdateEventH\x00\x42\x07\n\x05\x65vent"<\n\x0bTokenEffect\x12-\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x1d.concordium.v2.plt.TokenEvent"\x90\x01\n\x17TokenModuleRejectReason\x12,\n\x08token_id\x18\x01 \x01(\x0b\x32\x1a.concordium.v2.plt.TokenId\x12\x0c\n\x04type\x18\x02 \x01(\t\x12-\n\x07\x64\x65tails\x18\x03 \x01(\x0b\x32\x17.concordium.v2.plt.CborH\x00\x88\x01\x01\x42\n\n\x08_details"\xc6\x01\n\tCreatePLT\x12,\n\x08token_id\x18\x01 \x01(\x0b\x32\x1a.concordium.v2.plt.TokenId\x12\x37\n\x0ctoken_module\x18\x02 \x01(\x0b\x32!.concordium.v2.plt.TokenModuleRef\x12\x10\n\x08\x64\x65\x63imals\x18\x04 \x01(\r\x12:\n\x19initialization_parameters\x18\x05 \x01(\x0b\x32\x17.concordium.v2.plt.CborJ\x04\x08\x03\x10\x04"w\n\x14TokenCreationDetails\x12\x30\n\ncreate_plt\x18\x01 \x01(\x0b\x32\x1c.concordium.v2.plt.CreatePLT\x12-\n\x06\x65vents\x18\x02 \x03(\x0b\x32\x1d.concordium.v2.plt.TokenEventB=\n\x1a\x63om.concordium.grpc.v2.pltP\x01Z\x04./pb\xaa\x02\x16\x43oncordium.Grpc.V2.Pltb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "protocol_level_tokens_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals["DESCRIPTOR"]._serialized_options = (
        b"\n\032com.concordium.grpc.v2.pltP\001Z\004./pb\252\002\026Concordium.Grpc.V2.Plt"
    )
    _globals["_CBOR"]._serialized_start = 64
    _globals["_CBOR"]._serialized_end = 85
    _globals["_TOKENID"]._serialized_start = 87
    _globals["_TOKENID"]._serialized_end = 111
    _globals["_TOKENMODULEREF"]._serialized_start = 113
    _globals["_TOKENMODULEREF"]._serialized_end = 144
    _globals["_TOKENAMOUNT"]._serialized_start = 146
    _globals["_TOKENAMOUNT"]._serialized_end = 192
    _globals["_TOKENSTATE"]._serialized_start = 195
    _globals["_TOKENSTATE"]._serialized_end = 393
    _globals["_TOKENACCOUNTSTATE"]._serialized_start = 396
    _globals["_TOKENACCOUNTSTATE"]._serialized_end = 545
    _globals["_TOKENMODULEEVENT"]._serialized_start = 547
    _globals["_TOKENMODULEEVENT"]._serialized_end = 621
    _globals["_TOKENHOLDER"]._serialized_start = 623
    _globals["_TOKENHOLDER"]._serialized_end = 697
    _globals["_TOKENTRANSFEREVENT"]._serialized_start = 700
    _globals["_TOKENTRANSFEREVENT"]._serialized_end = 907
    _globals["_TOKENSUPPLYUPDATEEVENT"]._serialized_start = 909
    _globals["_TOKENSUPPLYUPDATEEVENT"]._serialized_end = 1029
    _globals["_TOKENEVENT"]._serialized_start = 1032
    _globals["_TOKENEVENT"]._serialized_end = 1355
    _globals["_TOKENEFFECT"]._serialized_start = 1357
    _globals["_TOKENEFFECT"]._serialized_end = 1417
    _globals["_TOKENMODULEREJECTREASON"]._serialized_start = 1420
    _globals["_TOKENMODULEREJECTREASON"]._serialized_end = 1564
    _globals["_CREATEPLT"]._serialized_start = 1567
    _globals["_CREATEPLT"]._serialized_end = 1765
    _globals["_TOKENCREATIONDETAILS"]._serialized_start = 1767
    _globals["_TOKENCREATIONDETAILS"]._serialized_end = 1886
# @@protoc_insertion_point(module_scope)
