# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kernel.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0ckernel.proto\x12\rconcordium.v2\"\x1f\n\x0e\x41\x63\x63ountAddress\x12\r\n\x05value\x18\x01 \x01(\x0c\"\x15\n\x04Memo\x12\r\n\x05value\x18\x01 \x01(\x0c\x42\x35\n\x16\x63om.concordium.grpc.v2P\x01Z\x04./pb\xaa\x02\x12\x43oncordium.Grpc.V2b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kernel_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\026com.concordium.grpc.v2P\001Z\004./pb\252\002\022Concordium.Grpc.V2'
  _globals['_ACCOUNTADDRESS']._serialized_start=31
  _globals['_ACCOUNTADDRESS']._serialized_end=62
  _globals['_MEMO']._serialized_start=64
  _globals['_MEMO']._serialized_end=85
# @@protoc_insertion_point(module_scope)
