import kernel_pb2 as _kernel_pb2
import types_pb2 as _types_pb2
import protocol_level_tokens_pb2 as _protocol_level_tokens_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor
